
{$A+,B-,D+,E+,F-,I+,L+,N-,O-,R+,S+,V+}

unit mnConfMa;

interface

uses
  Crt, Dos, m7Dec, m7Config, m7disk, m7UtilLo, m7time, m7utilg,
  mnDef, m7Msg, Version, m7FileIO, m7grplo, mnLog
  {$ifdef OS2} ,use32 {$endif};

var
  ThisKuerzel     : tKuerzel; { mu reichen! }
  TestConfigOnly  : Boolean;
  Debugging       : Boolean;
  OldBoxen        : ^ BoxenArray;
  OldGateWay      : ^ GateWayTyp;

procedure ClearConfig;

function ReadNewConfigFile1(const kz: string): Box;
procedure ReadNewConfigFile2(b: Box; const kz: string);

procedure ComputeConfig;

procedure ConfigThis;

implementation

type
  TelTransRec     =
    record
      Find,
      Repl            : string[20];
    end;
const
  TelTransMax     = 10;
var
  TelTrans        : array [1 .. TelTransMax] of ^ TelTransRec;
  WollteGruppe,
  BekamGruppe     : ^ BoxGruppenArray;
  Level,
  OnTop,
  BoxSort,
  TelSort         : array [Box] of Box;
  GrpNSort,
  GrpDSort        : array [NetzGruppe] of NetzGruppe;
  OnTopS,
  OldKz           : array [Box1] of ^ tKuerzel;
  Known,
  HaveConfig,
  ConfOk,
  Renamed         : array [Box] of Boolean;
  TelGeheim       : array [1..MaxNummern, Box] of Boolean;
  GateBox,
  GateCost { # Hops }
                  : array [Gateway1] of Box;
  OldNetGrpInfo   : NetGrpInfoTyp;
  NextGate        : Gateway1;
  LBL,
  SBI             : Text;
  MyGrps          : array [NetzGruppe] of ^ string;
  CnfValid        : Boolean;
  Domains         : array [1 .. 10] of DomainStr;
  NextDomain      : system.word;
  WriteSuccessMsg : Boolean;

procedure Warn(const s: string);
begin
  SysopLog(s);
end;

procedure Invalid(const s: string);
begin
  Warn(s);
  CnfValid := False;
end;

procedure FatalError(const s: string);
begin
  Warn(s);
  SysopProtExit('MausNet', 'MNConfig abgebrochen!');
  WriteLn(Log, 'MNConfig abgebrochen!'#7#7#7);
  Halt(255);
end;

procedure IncNextGate;
begin
  if NextGate = DefGateMax then
    FatalError('berlauf - zu viele Gates');
  Inc(NextGate);
end;

function GruppenNummer(NetNo: NetzGruppe; const lGrpName, nGrpName: string):
    tGruppe;
var
  c               : tGruppe;
begin
  { Finden wir sie unter der Netznummer? }
  c := 0;
  if NetNo <> 0 then
    repeat
      inc(c);
      GetGruppeDirect(c);
      if Gruppe.NetGrpNum = NetNo then begin
        GruppenNummer := c;
        Exit;
      end;
    until c = GrpMax;

  { mit dem angegebenen Namen suchen }
  c := 0;
  if lGrpName <> '' then
    repeat
      inc(c);
      if UmlautGleich(GruppenName(c), lGrpName) then begin
        GruppenNummer := c;
        Exit;
      end;
    until c = GrpMax;

  { vielleicht ist die Gruppe unter dem Netznamen zu finden? }
  c := 0;
  if nGrpName <> '' then
    repeat
      inc(c);
      if UmlautGleich(GruppenName(c), nGrpName) then begin
        GruppenNummer := c;
        Exit;
      end;
    until c = GrpMax;

  GruppenNummer := NoGrp;
end {GruppenNummer};

function IsSecret(GruppenNr: tGruppe): Boolean;
begin
  IsSecret := GrpIdx^[GruppenNr].Geheim;
end;

function SecretOk(GruppenNr: tGruppe): Boolean;
begin
  with GrpIdx^[GruppenNr] do
    SecretOk := not (Oeffentlich or (not Geheim) or SogarGaeste or
        WriteAccess);
end;

procedure MakeSecret(GruppenNr: tGruppe; DoIt: Boolean);
begin
  with GrpIdx^[GruppenNr] do
    if not SecretOk(GruppenNr) then begin
      Warn('Gruppe ' + Name^ + ' ist nicht geheim');
      if DoIt then begin
        Warn(' - wird korrigiert');
        GetGruppeDirect(GruppenNr);
        with Gruppe.Flags do begin
          Oeffentlich := False;
          Geheim := True;
          SogarGaeste := False;
          WriteAccess := False;
          Default := False;
        end;
        PutGruppeDirect(GruppenNr);
      end {if DoIt};
    end {if not SecretOk(GruppenNr)};
end {MakeSecret};

function ReadNewConfigFile1(const kz: string): Box;
var
  b               : Box;
  HaveConf        : Boolean;
  FileName        : string;
  fi              : Text;
  fx              : Text;
  Name            : string;
  Line            : string;
  got             : set of Char;
  bi              : BoxInfoRecTyp;
  i,j             : Longint;
  TelGeh          : array[1..MaxNummern] of Boolean;
  ok              : Boolean;
  OnTopStr        : string;
  port            : system.integer;
  what            : char;
  nr              : ^string;
  renameExp       : tKuerzel;
  newDate,
  oldDate         : longint;

  procedure WarnF(s: string);
  begin
    if (b = ThisBox) or (kz = ThisKuerzel) or HaveConf then
      Warn(FileName + ' - ' + s)
    else
      WriteLn(Log, FileName, ' - ', s);
  end;

  procedure Once;
  var
    CH              : Char;
  begin
    CH := UpCase(Line[1]);
    if CH in got then
      Invalid(FileName + ' - Zeile doppelt [' + CH + ']');
    Include(got, CH);
  end;

  procedure FixupNummer(var nr: string);
  var
    i : integer;
  begin
    for i := 2 to Length(nr) do
      if not IsDigit(nr[i]) then
        nr[i] := '-';
    i := TelTransMax;
    while (i > 0) and
      (TelTrans[i]^.Find <> Copy(nr, 1, Length(TelTrans[i]^.Find))) do
      Dec(i);
    if i > 0 then
      nr := TelTrans[i]^.Repl + copyR(nr, Length(TelTrans[i]^.Find) + 1);
  end;

  procedure DoRename(const o, n, ext: string);
  var
    old, new: string;
  begin
    old := o + ext;
    new := n + ext;
    if DateiDa(new) then
      SysopLog('Kann ' + old + ' nicht umbenennen, ' + new + ' existiert!')
    else begin
      SysopLog('Benenne ' + old + ' in ' + new + ' um.');
      IO(RenameFile(old, new));
    end;
  end;

  procedure StoreCollision(const neu, alt: tKuerzel);
  begin
    Warn('Vorrang hat '+neu+', MN-'+alt+'.EXP wird in .EX umbenannt.');
    renameExp := alt;
  end;

  procedure RenameCollision(const kz: tKuerzel);
  var
    old, new: string;
  begin
    old := MakeNameNeuK(kz) + '.exp';
    new := MakeNameNeuK(kz) + '.ex';
    if DateiDa(new) then
      IO(DelFile(new));
    SysOpLog('wg. Kollision: Benenne ' + old + ' in ' + new + ' um.');
    IO(RenameFile(old, new));
  end;

  function GetExpDate(const kz: tKuerzel): longint;
  var
    l  : longint;
    f  : file;
  begin
    assign(f, MakeNameNeuK(kz)+'.exp');
    {$i-}
    FileMode := fmReadAlone;
    reset(f);
    FileMode := 2;
    {$i+}
    if (ioresult <> 0) then
      GetExpDate := 0
    else begin
      GetFTime(f, l);
      close(f);
      GetExpDate := l;
    end;
  end;

var
  geheim  : boolean;
  p       : char;

begin {ReadNewConfigFile1}
  renameExp := '';
  CnfValid := True;
  Name := MakeNameNeuK(kz);
  FileName := Name + '.cnf';
  ok := OpenText(fi, fmReadAlone, FileName);
  Assign(fx, Name + '.exp');
  HaveConf := True;
  if not ok then begin { ok, versuchen wir ein Export-File }
    FileName := Name + '.exp';
    ok := OpenText(fi, fmReadAlone, FileName);
    Assign(fx, 'nul'); { Dann erzeugen wir natrlich keins }
    HaveConf := False;
    if not ok then
      Exit;
  end;
  Rewrite(fx);

  FillChar(bi, SizeOf(bi), 0);
  TelGeh[1] := True;
  bi.Nummer1 := '0'; { so ist T optional }
  bi.ConnTyp[1] := [];

  TelGeh[2] := True;
  bi.Nummer2 := '';
  bi.ConnTyp[2] := [];
  port := 0;

  b := 0;
  got := [];
  bi.Kuerzel := #13; { sonst knallt's bei fehlendem Krzel }
  while not Eof(fi) do
    with bi do begin
      ReadLn(fi, Line);
      Trim(Line);
      if Line <> '' then begin
        if Line[1] <> ';' then
          WriteLn(fx, Line);
        case Line[1] of
          ';': ; { Kommentar }
          '*':
            begin
              Once;
              Delete(Line, 1, 1);
              Trim(Line);
              UpString(Line);
              if (Length(Line) > High(Kuerzel)) or (Length(Line) < 1) then
                Invalid(FileName + ' - Illegales Krzel [*]');
              Kuerzel := Line;
            end;
          '-':
            begin
              Once;
              Delete(Line, 1, 1);
              Trim(Line);
              UpString(Line);
              if (Length(Line) > High(Kuerzel)) or (Length(Line) < 1) then
                Invalid(FileName + ' - Illegale TopBox [-]');
              OnTopStr := Line;
            end;
          '#':
            begin
              Once;
              Delete(Line, 1, 1);
              Trim(Line);
              Val(Line, i, err);
              FremdBox := i < 0;
              i := Abs(i);
              if i <= DefBoxenMax then
                b := i
              else
                err := 1;
              if err <> 0 then
                Invalid(FileName + ' - Illegale Boxnummer [#]');
            end {'#'};
          'N':
            begin
              Once;
              Delete(Line, 1, 1);
              Trim(Line);
              if (Length(Line) > High(Name)) or (Length(Line) < Length(
                  Kuerzel)) then
                Invalid(FileName + ' - Illegaler Boxname [N]');
              Name := Line;
            end;
          'n': Invalid(FileName + ' - Boxname mu public sein [n]');
          'T', 't':
            begin
              TelGeh[1] := Line[1] = 't';
              NrGeheim[1] := TelGeh[1];
              Once;
              Delete(Line, 1, 1);
              Trim(Line);
              if Length(Line) > 20 then
                Invalid(FileName + ' - Illegale primre Rufnummer');
              Nummer1 := Line;
              FixupNummer(Nummer1);
            end {'T','t'};
          'P', 'p':
            { [pP][#tk][nummer|ad] }
            begin
              what := upcase(line[2]);
              p := line[1];
              geheim := (p = 'p') or (line[2] = 't');
              Delete(Line, 1, 2);
              Trim(Line);
              case what of
                '#':
                  begin
                    port := value(line);
                    if not (port in [1..MaxNummern]) then begin
                      Invalid(FileName + ' - Ungltige Portnummer');
                      port := 0;
                    end
                    else begin
                      TelGeh[port] := geheim;
                      NrGeheim[port] := geheim;
                    end;
                  end;
                'K':
                  if port in [1..MaxNummern] then begin
                    if TelGeh[port] <> (p = 'p') then
                      Invalid(Filename + ' - P-Zeilen nicht konsistent gro/klein');
                    for i:=1 to length(line) do begin
                      j := pos(upcase(line[i]), cConnectTags);
                      if (j=0) then begin
                        Warn(Filename + ' - Illegaler Connection-Typ "' + line[i] + '"');
                        ConnTyp[port] := [];
                        break;
                      end
                      else begin
                        Include(ConnTyp[port], tConnectTypes(j-1));
                      end;
                    end;
                    if (ConnTyp[port] - [cIsdnX75, cAnalogV32bis] <> []) and
                       not (cAnalogV32bis in ConnTyp[port]) then
                      if ThisBox = b then
                        Warn(Filename + ' - "A" fehlt in PK-Zeile fr Port '+IntToStr(port));
                  end
                  else
                    Invalid(Filename + ' - PK ohne P#');
                'T':
                  if port in [1..MaxNummern] then begin
                    if TelGeh[port] <> (p = 'p') then
                      Invalid(Filename + ' - P-Zeilen nicht konsistent gro/klein');
                    if Length(Line) > 20 then
                      Invalid(FileName + ' - Illegale Rufnummer, Port'+IntToStr(port));
                    if (port=1) then
                      nr := @Nummer1
                    else
                      nr := @Nummer2;
                    nr^ := Line;
                    FixupNummer(nr^);
                  end
                  else
                    Invalid(Filename + '  - PT ohne P#');
                else
                  Invalid(Filename + ' - unbekannter P-Zeilen-Typ "'+what+'"');
              end;
            end;
          { Diverse Angaben, die erst spter bearbeitet werden
              - hier nur kurze Checks }
          'D': Once;
          'd': Invalid(FileName + ' - Hauptdomain mu public sein [d]');
          'Z', 'z': ;
          'G': ;
          'g': Invalid(FileName + ' - geheime Gates nicht untersttzt [g]');
          '%': ;
          'S', 's': ;
          'U': ;
          'u': Invalid(FileName + ' - User-Info mu public sein [u]');
          else WarnF('Unbekannte Zeile ignoriert [' + Line[1] + ']');
        end {case Line[1]};
      end {if Line<>''};
    end {with bi};

  { Auswerten }
  if not ('*' in got) then
    Invalid(FileName + ' - Krzel fehlt [*]');
  if not ('#' in got) then
    Invalid(FileName + ' - Netznummer fehlt [#]');
  if not ('N' in got) then
    Invalid(FileName + ' - Name fehlt [N]');

  for i := 1 to DefBoxenMax do
    if (i <> b) and (Boxen^[i].Kuerzel = bi.Kuerzel) then
      Invalid(FileName + ' - Boxname doppelt [*]');

  if CnfValid then begin
    if (ThisBox = 0) and (ThisKuerzel = bi.Kuerzel) then
      ThisBox := b;
    if OldBoxen^[b].Kuerzel <> bi.Kuerzel then begin
      Warn('Box ' + IntToStr(b) + ' ' + bi.Kuerzel + ' ' + bi.Name +
          ' neu im Netz');
      if OldBoxen^[b].Kuerzel <> '' then begin
        newDate := GetExpDate(bi.Kuerzel);
        oldDate := GetExpDate(OldBoxen^[b].Kuerzel);
        { eDate ist 0, wenn's das neue .EXP }
        if (newDate <> 0) and (oldDate <> 0) then begin
          with OldBoxen^[b] do
            Warn('Kollision mit Box '+IntToStr(b)+' '+Kuerzel+' ' +Name+'!');
          if (newDate > oldDate) then
            StoreCollision(bi.Kuerzel, OldBoxen^[b].Kuerzel)
          else begin
            StoreCollision(OldBoxen^[b].Kuerzel, bi.Kuerzel);
            bi := OldBoxen^[b];
          end;
        end
        else { eine Umbenennung }
          with OldBoxen^[b] do
            warn('umbenannt, hie frher ' + Kuerzel + ' ' + Name + '.');
      end
      else
        Warn('Gratulation!');
    end;

    Boxen^[b] := bi;
    if '-' in got then begin
      OnTop[b] := DefBoxenMax;
      OnTopS[b]^ := OnTopStr;
    end;
    for i:=1 to MaxNummern do
      TelGeheim[i][b] := TelGeh[i];

    if b = ThisBox then begin { Domains einsammeln! }
      Reset(fi);
      while not Eof(fi) do begin
        ReadLn(fi, Line);
        if (Line <> '') and (Pos(Line[1], 'DZz') > 0) then begin
          Domains[NextDomain] := UpStr(Copy(Line, 2, 255));
          WriteLn(Log, 'Lokale Domain[', NextDomain, '] = ', Domains[
              NextDomain]);
          Inc(NextDomain);
        end;
      end;
    end {if b=ThisBox};
  end {if CnfValid}
  else begin
    Warn(FileName + ' - ignoriert');
    if b = 0 then begin
      if '*' in got then
        Line := bi.Kuerzel
      else
        Line := kz;
      for i := 1 to DefBoxenMax do
        if OldBoxen^[i].Kuerzel = Line then
          b := i;
    end;
    if b <> 0 then begin
      Boxen^[b] := OldBoxen^[b];
      if '-' in got then begin
        OnTop[b] := DefBoxenMax;
        OnTopS[b]^ := OnTopStr;
      end;
      Warn('Verwende stattdessen alte Daten von ' + IntToStr(b) + ' ' + Boxen^
          [i].Kuerzel);
      for i:=1 to MaxNummern do
        TelGeheim[i][b] := True; { sicher ist sicher ... }
    end;
  end {if CnfValid else};

  if b > 0 then
    ConfOk[b] := CnfValid;

  if b > 0 then begin
    Write(Log, 'Box ', b, ' ist ', Boxen^[b].Name);
    WriteLnLog;
    Known[b] := True;
    HaveConfig[b] := HaveConf;
  end;

  if b > MaxBoxen then
    MaxBoxen := b;

  Close(fi);
  Close(fx);

  if CnfValid then begin
    if (kz <> Boxen^[b].Kuerzel) and (kz <> renameExp) then begin
      Renamed[b] := True;
      New(OldKz[b]);
      OldKz[b]^ := kz;
      if not TestConfigOnly then begin
        FileName := MakeNameNeuK(Boxen^[b].Kuerzel);
        if HaveConf then
          DoRename(Name, Filename, '.cnf');
        DoRename(Name, Filename, '.exp');
        SysopLog('.CNF/.EXP umbenannt: ' + Name + '.* -> ' + FileName + '.*');
      end;
    end {if CnfValid and(kz<>Boxen^[b].Kuer...};
    if (not TestConfigOnly) and (renameExp <> '') then
      RenameCollision(renameExp);
  end;

  ReadNewConfigFile1 := b;
end {ReadNewConfigFile1};

procedure ReadNewConfigFile2(b: Box; const kz: string);
var
  i               : Longint;
  FileName        : string;
  fi              : Text;
  SBItemp,
  LBLtemp         : Text;
  num1,
  num2            : string;
  Line            : string;
  FirstGate       : Gateway1;
  Alias           : file of tNetAlias;

  procedure AddAlias(b: Box; var line: string);
  var
    i       : integer;
    netalias: tNetAlias;
  begin
    if TestConfigOnly then
      exit;

    fillchar(netalias, sizeof(netalias), 0);
    netalias.boxnr := b;
    netalias.domain := UpStr(copy(line, 2, 255));

    if netalias.domain = UpStr(ExMausSubOrg) then begin
      if (b=ThisBox) then
        SysOpLog(ExMausSubOrg + ' gibt''s nicht mehr, bitte auskonfigurieren!');
      exit;
    end;
    if netalias.domain = UpStr(MausDe) then
      exit;

    if netalias.domain[1] = '.' then
      netalias.domain := UpStr(BoxName(b)) + netalias.domain;

    write(alias, netalias);
    Write(Log, 'Alias "', netalias.domain, '" fr ', BoxName(b), ' eingetragen');
    WriteLnLog;
  end;

  procedure WarnF(s: string);
  begin
    if (b = ThisBox) or (kz = ThisKuerzel) or HaveConfig[b] then
      Warn(FileName + ' - ' + s)
    else
      WriteLn(Log, FileName, ' - ', s);
  end;

  procedure PutGateInfo;
  var
    g,
    g2              : Integer;
    GDone           : array [Gateway1] of Boolean;
  begin
    if FirstGate <> NextGate then begin
      WriteLn(LBLtemp, 'Gateway(s):');
      FillChar(GDone, SizeOf(GDone), False);
      for g := FirstGate to NextGate - 1 do
        if not GDone[g] then begin
        { Mehrfachgates erklren }
          for g2 := FirstGate to NextGate - 1 do
            if GateWay^[g2].GateDist = GateWay^[g].GateDist then begin
              GDone[g2] := True;
              Write(LBLtemp, GateWay^[g2].GateName, ' ');
            end;
          WriteLn(LBLtemp);
        end;
      WriteLn(LBLtemp);
    end {if FirstGate<>NextGate};
  end {PutGateInfo};

  procedure AllowGate(const Name: string; b: Byte; const who: string);
  begin
    Write(Log, 'Gate ', name, ' in ', BoxName(b), ' fr ', who, ' erlaubt');
    WriteLnLog;
  end;

var
  localno         : Integer;
  Name            : string;
  other           : Gateway0;
  g               : Gateway1;
  p,
  q               : Byte;
  ok              : Boolean;
begin {ReadNewConfigFile2}
  if not Known[b] then
    Exit;
  Write(Log, b, ' = ', BoxName(b));
  WriteLnLog;
  if TestConfigOnly then begin
    Assign(SBItemp, 'nul');
    Assign(LBLtemp, 'nul');
    Assign(Alias, 'nul');
  end
  else begin
    Assign(SBItemp, ConfPath + 'Temp.SIn');
    Assign(LBLtemp, ConfPath + 'Temp.Inf');
    Assign(Alias, MausNetPath + AliasFile);
  end;

  Rewrite(SBItemp);
  Rewrite(LBLtemp);
  {$i-}
  reset(Alias);
  {$i+}
  if NOT TestConfigOnly then begin (* OS2 mag kein Seek in 'NUL' *)
    if IOresult <> 0 then          (* ist eh unschn *)
      rewrite(Alias);
    seek(Alias, filesize(Alias));
  end;

  with Boxen^[b] do begin
    if TelGeheim[1][b] then
      num1 := '[nicht ffentlich]'
    else
      num1 := Nummer1;
    if (Nummer2<>'') and TelGeheim[2][b] then
      num2 := '[nicht ffentlich]'
    else
      num2 := Nummer2;

    Write(LBLtemp, b);
    WriteLn(LBLtemp, ' ', Name, '   (Krzel: ', Kuerzel, ')');
    if b >= 10 then
      Write(LBLtemp, '-');
    if b >= 100 then
      Write(LBLtemp, '-');
    Write(LBLtemp, '--');
    for i := 1 to Length(Name) do
      Write(LBLtemp, '-');
    WriteLn(LBLtemp);
    WriteLn(LBLtemp);
    WriteLn(LBLtemp, 'Nummer:    ', num1);
    if (num2 <> '') then
      WriteLn(LBLtemp, '2. Nummer: ', num2);
    WriteLn(LBLtemp);
  end {with Boxen^[b]};

  GetBoxGrpRec(b, WillGruppe^[b]);

  FirstGate := NextGate;

  if ConfOk[b] then begin { wir knnen das .EXP nehmen }
    if Renamed[b] and TestConfigOnly then
      FileName := MakeNameNeuK(OldKz[b]^) + '.exp'
    else
      FileName := MakeNameNeu(b) + '.exp';
    IO(OpenText(fi, fmReadAlone, FileName));
    while not Eof(fi) do begin
      ReadLn(fi, Line);
      Trim(Line);
      if Line <> '' then
        case Line[1] of
          'G':
            begin
              p := Pos(',', Line);
              q := Pos(',', Copy(Line, p + 1, 255));
              if (p = 0) or (q = 0) or (p + q = Length(Line)) then
                WarnF('Illegale G-Zeile: ' + Line + ' (ignoriert)')
              else begin
                Name := Copy(Line, 2, p - 2);
                Trim(Name);
                UpString(Name);
                Val(Copy(Line, p + 1, q - 1), i, err);
                if (i > - 1000) and (i < NetzMitt) and (err = 0) then begin
                  localno := i;
                  { ok, soweit, so gut ... }
                  Delete(Line, 1, p + q);
                  for i := Length(Line) downto 1 do
                    if Line[i] = ' ' then
                      Delete(Line, i, 1);
                  UpString(Line);
                  Line := ',' + Line + ',';
                  ok := Pos(',*,', Line) > 0;
                  if ok then AllowGate(name, b, 'alle');
                  if not ok then begin
                    ok := Pos(',' + ThisKuerzel + ',', Line) > 0;
                    if ok then AllowGate(name, b, 'uns');
                  end;
                  if not ok then begin
                    Delete(Line, 1, 1);
                    repeat
                      p := Pos(',', Line);
                      if p > 0 then begin
                        if p > 1 then
                          for i := 1 to NextDomain - 1 do
                            if Copy(Line, 1, p) = (Domains[i] + ',') then begin
                              ok := True;
                              AllowGate(name, b, Domains[i]);
                            end;
                        Delete(Line, 1, p);
                      end;
                    until p = 0;
                  end {if not ok};
                  if ok then begin
                    with GateWay^[NextGate] do begin
                      GateName := Name;
                      GateDist := localno;
                    end;
                    GateBox[NextGate] := b;
                    IncNextGate;
                  end;
                end {if(i>-1000)and(i<NetzMitt)and(err=0)}
                else begin
                  localno := MaxInt;
                  WarnF('Illegale Gatewaynummer: ' + Line + ' (ignoriert)');
                end;
              end {if(p=0)or(q=0)or(p+q=Length(Line))...};
            end {'G'};
          'S', 's', '%': WriteLn(SBItemp, Line);
          'U': WriteLn(LBLtemp, Copy(Line, 2, 255));
          'D', 'z', 'Z': AddAlias(b, line);
        end {case Line[1]};
    end {while not Eof(fi)};
    Close(fi);
    WriteLn(LBLtemp);
  end {if ConfOk[b]}
  else begin { kein heiles .EXP :-( }
    for g := 1 to DefGateMax do begin
      if OldGateWay^[g].GateDist = b then begin
        GateWay^[NextGate] := OldGateWay^[g];
        { Kaputte GateDist - kann ich nicht ndern! }
        GateBox[NextGate] := b;
        IncNextGate;
      end;
    end;
    WriteLn(SBItemp, '>>> .EXP kaputt!');
    WriteLn(LBLtemp, '>>> Beschreibung zu dieser Box leider defekt!');
    WriteLn(LBLtemp);
  end {if ConfOk[b]else};

  PutGateInfo;

  if not TestConfigOnly then begin
    if CnfValid then begin
      ReOpenText(LBLtemp, fmReadAlone);
      while not Eof(LBLtemp) do begin
        ReadLn(LBLtemp, Line);
        WriteLn(LBL, Line);
      end;
      WriteLn(SBI);
      with Boxen^[b] do begin
        Write(SBI, '** ', b: 3, ' ', Kuerzel, '': 9 - Length(Kuerzel));
        Write(SBI, Name, '': 30 - Length(Name));
        Write(SBI, Nummer1);
        if (Nummer2 <> '') then
          WriteLn(SBI, ', ', Nummer2)
        else
          WriteLn(SBI);
        WriteLn(SBI);
      end;
      ReOpenText(SBItemp, fmReadAlone);
      while not Eof(SBItemp) do begin
        ReadLn(SBItemp, Line);
        WriteLn(SBI, Line);
      end;
    end {if CnfValid};
    Close(LBLtemp);
    Close(SBItemp);
    Erase(LBLtemp);
    Erase(SBItemp);
  end {if not TestConfigOnly}
  else begin
    Close(LBLtemp);
    Close(SBItemp);
  end;
  close(Alias);
end {ReadNewConfigFile2};

procedure ClearConfig;
var
  c               : Word;
  s,
  BoxN            : string;
begin
{ Tabellen initialisieren }
  FillChar(Renamed, SizeOf(Renamed), False);
  for c := 0 to DefGrpMax do
    GrpToNetGrp[c] := NoGrp;
  for c := 0 to DefGrpMax do begin
    New(NetGrpInfo[c]);
    FillChar(NetGrpInfo[c]^, SizeOf(NetGrpInfo[c]^), 0);
    with NetGrpInfo[c]^ do begin
      ToGrp := NoGrp;
      FillChar(ToBox, SizeOf(ToBox), NichtImNetz);
    end;
  end;
  FillChar(Boxen^, SizeOf(Boxen^), 0);
  for c := 1 to DefBoxenMax do begin
    OnTop[c] := 0;
    Level[c] := 0;
    New(OnTopS[c]);
    OnTopS[c]^ := '';
    HaveConfig[c] := False;
    ConfOk[c] := False;
    Boxen^[c].FremdBox := False;
  end;
  Level[0] := 0;
  FillChar(WillGruppe^, SizeOf(WillGruppe^), 0);
  FillChar(BekommtGruppe^, SizeOf(BekommtGruppe^), 0);
  FillChar(Known, SizeOf(Known), False);
  FillChar(TelGeheim, SizeOf(TelGeheim), False);
  New(GateWay);
  FillChar(GateWay^, SizeOf(GateWay^), 0);
  FillChar(GateBox, SizeOf(GateBox), 0);
  FillChar(GateCost, SizeOf(GateCost), 0);
  NextGate := 1;
  FillChar(DirectToBox, SizeOf(DirectToBox), 0);
  FillChar(Domains, SizeOf(Domains), 0);
  NextDomain := 1;
  New(OldGateWay);
  { die alte Konfiguration holen }
  if DatEiDa(MausNetPath + MausNetDatFile) then begin
    NetConfigEinlesen(False, NilGrpToNetGrp, OldNetGrpInfo, OldBoxen^,
        OldGateWay^, NilDirectToBox, GrpInfo, WollteGruppe^, BekamGruppe^);
  end
  else begin
    ThisBox := 0;
    MaxDirectBox := 0;
    NetConfigSchreiben(False); { wir mssen so ein Teil haben }
    NetConfigEinlesen(False, NilGrpToNetGrp, NilNetGrpInfo, NilBoxen,
        NilGateWay, NilDirectToBox, GrpInfo, NilBoxGruppe, NilBoxGruppe);
    for c := 0 to DefGrpMax do begin
      New(OldNetGrpInfo[c]);
      FillChar(OldNetGrpInfo[c]^, SizeOf(OldNetGrpInfo[c]^), NichtImNetz);
    end;
    { nur .ToBox wird gebraucht! }
    FillChar(OldBoxen^, SizeOf(OldBoxen^), 0);
    for c := 1 to DefBoxenMax do
      OldBoxen^[c].Direct := DefDirectBoxMax;
    FillChar(WollteGruppe^, SizeOf(WollteGruppe^), 0);
    FillChar(BekamGruppe^, SizeOf(BekamGruppe^), 0);
    FillChar(OldGateWay^, SizeOf(OldGateWay^), 0);
  end {if DatEiDa(MausNetPath+MausNetDatF...};
  MaxBoxen := 0;
  for c := 0 to DefGrpMax do
    if GrpInfo[c] = nil then begin
      New(GrpInfo[c]);
      FillChar(GrpInfo[c]^, SizeOf(GrpInfo[c]^), 0);
    end;
  if TestConfigOnly then
    Exit;
  Assign(LBL, MausNetPath + 'LBoxList.Inf');
  Rewrite(LBL);
  Assign(SBI, MausNetPath + 'SBoxInf.Inf');
  Rewrite(SBI);
end {ClearConfig};

function NetCost(b1, b2: Box): Word;
var
  c               : Word;

  function Fern(b: Box): Boolean;
  { ist die TopBox von b ein Ferngesprch? }
  var
    bn,
    on              : string[20];
    O               : Box;
    j,
    k               : Byte;
  begin
    bn := Boxen^[b].Nummer1;
    O := OnTop[b];
    on := Boxen^[O].Nummer1;
    j := 1;
    k := 1;
    while (bn[j] = on[j]) and (j < Length(bn)) and (j < Length(on)) do begin
      if bn[j] = '-' then begin
        Fern := False;
        Exit;
      end;
      Inc(j);
    end;
    Fern := True
  end {Fern};

begin {NetCost}
  c := 0;
  while Level[b1] > Level[b2] do begin
    if Fern(b1) then
      Inc(c, 10)
    else
      Inc(c);
    b1 := OnTop[b1];
  end;
  while Level[b2] > Level[b1] do begin
    if Fern(b2) then
      Inc(c, 10)
    else
      Inc(c);
    b2 := OnTop[b2];
  end;
  while b1 <> b2 do begin
    if Fern(b1) then
      Inc(c, 10)
    else
      Inc(c);
    if Fern(b2) then
      Inc(c, 10)
    else
      Inc(c);
    b1 := OnTop[b1];
    b2 := OnTop[b2];
  end {while b1<>b2};
  NetCost := c;
end {NetCost};
{$f+}

function qscBox(before, after: Integer): Boolean;
begin
  qscBox := Boxen^[BoxSort[before]].Kuerzel <= Boxen^[BoxSort[after]].Kuerzel
end;

procedure qssBox(a, b: Integer);
var
  c               : Box;
begin
  c := BoxSort[a];
  BoxSort[a] := BoxSort[b];
  BoxSort[b] := c;
end;

function qscTel(before, after: Integer): Boolean;
var
  bef,
  aft             : string;
begin
  bef := Boxen^[TelSort[before]].Nummer1;
  aft := Boxen^[TelSort[after]].Nummer1;
  qscTel := (bef < aft) or ((bef = aft) and (TelSort[before] <= TelSort[after]
      ));
end;

procedure qssTel(a, b: Integer);
var
  c               : Box;
begin
  c := TelSort[a];
  TelSort[a] := TelSort[b];
  TelSort[b] := c;
end;

function qscGrpN(before, after: Integer): Boolean;
begin
  qscGrpN := UmlautVWandelF(GrpInfo[GrpNSort[before]]^.Name) <=
      UmlautVWandelF(GrpInfo[GrpNSort[after]]^.Name);
end;

procedure qssGrpN(a, b: Integer);
var
  c               : NetzGruppe;
begin
  c := GrpNSort[a];
  GrpNSort[a] := GrpNSort[b];
  GrpNSort[b] := c;
end;

function qscGrpD(before, after: Integer): Boolean;
begin
  qscGrpD := (GrpInfo[GrpDSort[before]]^.LastEdit < GrpInfo[GrpDSort[after]]^.
      LastEdit) or ((GrpInfo[GrpDSort[before]]^.LastEdit = GrpInfo[GrpDSort[
      after]]^.LastEdit) and ((GrpInfo[GrpDSort[before]]^.LastEditTime <
      GrpInfo[GrpDSort[after]]^.LastEditTime) or ((GrpInfo[GrpDSort[before]]^.
      LastEditTime = GrpInfo[GrpDSort[after]]^.LastEditTime) and (GrpDSort[
      before] <= GrpDSort[after]))));
end;

procedure qssGrpD(a, b: Integer);
var
  c               : NetzGruppe;
begin
  c := GrpDSort[a];
  GrpDSort[a] := GrpDSort[b];
  GrpDSort[b] := c;
end;
{$f-}

procedure ComputeConfig;

  procedure GenNetzPlan;
  var
    f               : Text;
    Level           : Byte;
    g,
    g2,
    n               : Gateway0;
    GDone           : array [Gateway1] of Boolean;
    count           : array [Box] of integer;
    b               : Box;

    function BoxClients(UnterBox: Box): integer;
    var
      b: box;
    begin
      BoxClients := 0;
      for b := 1 to DefBoxenMax do begin
        if Known[b] and (OnTop[b] = UnterBox) then begin
          inc(count[UnterBox], 1 + BoxClients(b));
          BoxClients := count[UnterBox];
        end;
      end;
    end;

    function Calc(UnterBox: Box): boolean;
    var
      b,
      g,
      g2,
      n               : Gateway0;
      first           : Boolean;
    begin
      Calc := false;
      Inc(Level);
      first := True;
      for b := 1 to DefBoxenMax do begin
        if Known[b] and (OnTop[b] = UnterBox) then begin
          Calc := true;
          if first then
            first := False
          else
            Write(f, '': (Level - 1) * (KuerzelLen+4) - 4);
          if Level > 1 then
            Write(f, ' -- ');
          Write(f, AlignL(Boxen^[b].Kuerzel, KuerzelLen));
          if not Calc(b) then
            WriteLn(f);
        end {if Known[b]and(OnTop[b]=UnterBox)};
      end {for b:=1 to DefBoxenMax};
      for g := 1 to NextGate - 1 do begin
        if not GDone[g] and (GateBox[g] = UnterBox) then begin
          { Mehrfachgates nur einmal nennen }
          n := 0;
          for g2 := 1 to NextGate - 1 do
            if (GateBox[g2] = UnterBox) and
               (GateWay^[g2].GateDist = GateWay^[g].GateDist) then begin
              Inc(n);
              GDone[g2] := True;
            end;
          Calc := true;
          if first then
            first := False
          else
            Write(f, '': (Level - 1) * (KuerzelLen+4) - 4);
          if Level > 1 then
            Write(f, ' -- ');
          if n <> 1 then
            WriteLn(f, '[', Boxen^[UnterBox].Kuerzel, ' G',
              GateWay^[g].GateDist, ']')
          else
            WriteLn(f, '[', GateWay^[g].GateName, ']');
        end {if not GDone[g]and(GateBox[g]=Unte...};
      end {for g:=1 to NextGate-1};
      Dec(Level);
    end {Calc};

  begin {GenNetzPlan}
    if TestConfigOnly then
      Exit;
    WriteLn(Log, '[GenNetzPlan]');
    FillChar(count, sizeof(count), 0);
    FillChar(GDone, SizeOf(GDone), False);
    Assign(f, MausNetPath + 'NETZPLAN.INF');
    Rewrite(f);
    Level := 0;
    WriteLn(f, 'Netzplan des MausNet');
    WriteLn(f, '--------------------');
    BoxClients(0);
    Calc(0);
    WriteLn(f);
    FillChar(GDone, SizeOf(GDone), False);
    for g := 1 to NextGate - 1 do
      if not GDone[g] then begin
      { Mehrfachgates erklren }
        n := 0;
        for g2 := 1 to NextGate - 1 do
          if (GateBox[g2] = GateBox[g]) and
             (GateWay^[g2].GateDist = GateWay^[g].GateDist) then begin
            Inc(n);
            GDone[g2] := True;
          end;
        if n <> 1 then begin
          Write(f, '[', Boxen^[GateBox[g]].Kuerzel, ' G',
            GateWay^[g].GateDist, '] =');
          for g2 := 1 to NextGate - 1 do
            if (GateBox[g2] = GateBox[g]) and
               (GateWay^[g2].GateDist = GateWay^[g].GateDist) then
              Write(f, ' ', GateWay^[g2].GateName);
          WriteLn(f);
        end;
      end {if not GDone[g]};
    Close(f);
  end {GenNetzPlan};

  procedure GenNetMap;
  var
    i,
    j               : Box;
    k               : NetzGruppe;
    f               : Text;
    first           : Boolean;
  begin
    WriteLn(Log, '[GenNetMap]');
    Assign(f, MausNetPath + 'NetMap.Txt');
    Rewrite(f);
    for i := 1 to MaxBoxen do
      if Known[i] then begin
        WriteLn(f, i, ' ', Boxen^[i].Name, ' ', Boxen^[i].Kuerzel);
        if OnTop[i] = 0 then
          WriteLn(f, '  Zentralknoten')
        else
          WriteLn(f, '  Unter ', OnTop[i]);
        for j := 1 to MaxBoxen do
          if Known[j] then
            if OnTop[j] = i then
              WriteLn(f, '  ber Knoten ', j);
        for k := 0 to DefGrpMax do begin
          if IsInGSet(k, WillGruppe^[i]) then
            WriteLn(f, '  Kennt Gruppe ', k, ' [', GrpInfo[k]^.Name, ']');
          first := True;
          for j := 1 to MaxBoxen do
            if Known[j] then
              if IsInGSet(k, BekommtGruppe^[i]) and IsInGSet(k, BekommtGruppe^
                  [j]) and ((OnTop[i] = j) or (OnTop[j] = i)) then
                if first then begin
                  Write(f, '  Tauscht Gruppe ', k, ' mit ', j);
                  first := False
                end
                else
                  Write(f, ', ', j);
          if not first then
            WriteLn(f);
        end {for k:=0 to DefGrpMax};
      end {if Known[i]};
    Close(f);
  end {GenNetMap};

  procedure FixNetwork;
  var
    TopBox,
    i,
    j               : Box;
  begin
    WriteLn(Log, '[FixNetwork]');
    if not Known[ThisBox] then begin
      FatalError('Lokale Konfiguration nicht gefunden!');
    end;
    for i := 1 to MaxBoxen do begin
      if Known[i] and (OnTop[i] = DefBoxenMax) then begin
        for j := 1 to MaxBoxen do
          if Known[j] and (OnTopS[i]^ = Boxen^[j].Kuerzel) then
            OnTop[i] := j;
        if OnTop[i] = DefBoxenMax then
          Warn(Boxen^[i].Name + ' - Topbox nicht gefunden');
      end;
      if ((i = ThisBox) or (Boxen^[i].FremdBox and (OnTop[i] = ThisBox))) and
          not HaveConfig[i] then
        FatalError(Boxen^[i].Name + ' - Kein .CNF fr diese Box gefunden!!');
      if (i <> ThisBox) and HaveConfig[i] and not (Boxen^[i].FremdBox and (
          OnTop[i] = ThisBox)) then
        FatalError('Darf fr Box ' + Boxen^[i].Name + ' (' + IntToStr(i) +
            ') kein .CNF haben!!!');
    end {for i:=1 to MaxBoxen};
    TopBox := ThisBox; { Wenn wir keine andere TopBox finden }
    for i := MaxBoxen downto 1 do
      if Known[i] and (OnTop[i] = 0) then
        TopBox := i;
    while (OnTop[TopBox] <> 0) and Known[OnTop[TopBox]] do
      TopBox := OnTop[TopBox];
    if OnTop[TopBox] <> 0 then begin
      Warn('Kann TopBox nicht finden, da ' + IntToStr(OnTop[TopBox]) + ' [' +
          OnTopS[TopBox]^ + '] ber ' + Boxen^[TopBox].Name + ' (' + IntToStr(
          TopBox) + ') fehlt!');
      OnTop[TopBox] := 0;
    end;
    WriteLn(Log, 'TopBox=', Boxen^[TopBox].Name, '(', TopBox, ')');
    for i := 1 to MaxBoxen do
      if (i <> TopBox) and Known[i] then begin
        if OnTop[i] = 0 then begin
          Warn('Alternative TopBox gefunden: ' + Boxen^[i].Name + '(' +
              IntToStr(i) + ')');
          Warn('- wird unter ' + Boxen^[TopBox].Name + '(' + IntToStr(TopBox)
              + ') gehngt');
          OnTop[i] := TopBox;
        end
        else
          if not Known[OnTop[i]] then begin
            Warn(IntToStr(OnTop[i]) + ' [' + OnTopS[i]^ + '] ber ' + Boxen^[
                i].Name + ' (' + IntToStr(i) + ') fehlt');
            Warn('- wird unter ' + Boxen^[TopBox].Name + '(' + IntToStr(
                TopBox) + ') gehngt');
            OnTop[i] := TopBox;
          end;
      end {if(i<>TopBox)and Known[i]};
  end {FixNetwork};

  procedure CalcLevels;

    function LevelOf(b: Box): Box;
    var
      b2              : Box;
    begin
      if b = 0 then
        LevelOf := 0
      else
        if Level[b] <> 0 then
          if Level[b] = DefBoxenMax then begin
            b2 := b;
            Write(Log, Boxen^[b2].Name, '->');
            repeat
              b2 := OnTop[b2];
              Write(Log, Boxen^[b2].Name, '->');
            until b2 = b;
            WriteLn(Log, '...');
            FatalError('Schleife im Netz bei ' + Boxen^[b].Name)
          end
          else
            LevelOf := Level[b]
        else begin
          Level[b] := DefBoxenMax;
          LevelOf := LevelOf(OnTop[b]) + 1;
          Level[b] := 0;
        end {if Level[b]<>0 else};
    end {LevelOf};

  var
    i               : Box;
  begin {CalcLevels}
    WriteLn(Log, '[CalcLevels]');
    for i := 1 to MaxBoxen do
      if Known[i] then begin
        Level[i] := LevelOf(i);
        Write(Log, '#', i, ' ', Boxen^[i].Name, ' Level=', Level[i]);
        WriteLnLog;
      end;
  end;

  procedure CalcGruppen;
  type
    gset            = tGruppenSet;

    function PassUp(b: Box; var g: gset): Box;
    var
      O               : Box;
      g2              : gset;
    begin
      O := OnTop[b];
      PassUp := O;
      if IsEmptyGSet(g) then
        Exit;
      g2 := BekommtGruppe^[b];
      IntersectGSet(g2, BekommtGruppe^[O]);
      SubtractGSet(g, g2);
      AddGSet(BekommtGruppe^[b], g);
      AddGSet(BekommtGruppe^[O], g);
    end {PassUp};

  var
    i,
    j,
    l,
    m               : Box;
    k               : NetzGruppe;
    gsl,
    gsm             : gset;
  begin {CalcGruppen}
    WriteLn(Log, '[CalcGruppen]');
    { Fremdboxen kriegen nur, was die Anschlu-MAUS kriegt }
    for i := 1 to MaxBoxen do
      if Known[i] then begin
        j := i;
        while Boxen^[j].FremdBox do
          j := OnTop[j];
        IntersectGSet(WillGruppe^[i], WillGruppe^[j]);
      end;
    { Das unterstellt, da die Netzzentrale keine Fremdbox ist :-) }

    for i := 1 to MaxBoxen - 1 do
      if Known[i] then
        for j := i + 1 to MaxBoxen do
          if Known[j] then begin
            gsl := WillGruppe^[i];
            IntersectGSet(gsl, WillGruppe^[j]);
            if not IsEmptyGSet(gsl) then begin
              gsm := gsl;
              l := i; m := j;
              while Level[l] > Level[m] do
                l := PassUp(l, gsl);
              while Level[m] > Level[l] do
                m := PassUp(m, gsm);
              while l <> m do begin
                l := PassUp(l, gsl);
                m := PassUp(m, gsm);
              end;
            end {if not IsEmptyGSet(gsl)}
          end {if Known[j]};
  end {CalcGruppen};

  procedure CalcGateways;
  var
    i,
    j               : Box;
    g,
    g2              : Gateway0;
    c               : Box;
    s               : string;
    n               : Integer;
    w               : Word;
    Swap,
    Copy            : Boolean;
    d               : word;
  begin {CalcGateways}
    WriteLn(Log, '[CalcGateways]');

    with GateWay^[NextGate] do begin
      GateName := UpStr(MausDe);
      GateDist := NetzMitt;
    end;
    GateBox[NextGate] := ThisBox;
    IncNextGate;

    { das ist jetzt wirklich obsolet:
    with GateWay^[NextGate] do begin
      GateName := MausSubOrg;
      GateDist := NetzMitt;
    end;
    GateBox[NextGate] := ThisBox;
    IncNextGate;
    }

    { experimentell: Gates fr eigene Domains eintragen }
    for d := 1 to NextDomain - 1 do begin
      with Gateway^[NextGate] do begin
        GateName := Domains[d];
        GateDist := NetzMitt;
      end;
      GateBox[NextGate] := ThisBox;
      IncNextGate;
    end;

    for g := 1 to NextGate - 1 do begin { liest NextGate nur am Anfang }
      GateCost[g] := NetCost(ThisBox, GateBox[g]) + 1; { > Rerouted }
      GetConfigString(s, 'Route' + GateWay^[g].GateName, '');
      c := 0;
      if s <> '' then begin
        for i := 1 to MaxBoxen do
          if s = Boxen^[i].Kuerzel then
            c := i;
        if c = 0 then
          FatalError('Illegale Box "' + s + '" in Route' + GateWay^[g].
              GateName);
      end;
      if c <> 0 then begin
        GateWay^[NextGate] := GateWay^[g];
        GateCost[NextGate] := 0; { bestmglich }
        GateBox[NextGate] := c;
        GateWay^[NextGate].GateDist := - 999;
        IncNextGate;
        WriteLn(Log, 'Gateway fr ', GateWay^[g].GateName, ' geroutet nach ',
            s);
      end;
    end {for g:=1 to NextGate-1};
    { sortieren - bubbeln reicht }
    for g := 1 to NextGate - 2 do
      for g2 := g + 1 to NextGate - 1 do begin
        Swap := False;
        if GateCost[g] > GateCost[g2] then
          Swap := True
        else
          if GateCost[g] = GateCost[g2] then
            if GateBox[g] > GateBox[g2] then
              Swap := True
            else
              if GateBox[g] = GateBox[g2] then
                if GateWay^[g].GateName > GateWay^[g2].GateName then
                  Swap := True;
        if Swap then begin
          s := GateWay^[g].GateName; GateWay^[g].GateName := GateWay^[g2].
              GateName; GateWay^[g2].GateName := s;
          n := GateWay^[g].GateDist; GateWay^[g].GateDist := GateWay^[g2].
              GateDist; GateWay^[g2].GateDist := n;
          w := GateCost[g]; GateCost[g] := GateCost[g2]; GateCost[g2] := w;
          c := GateBox[g]; GateBox[g] := GateBox[g2]; GateBox[g2] := c;
        end;
      end {for g2:=g+1 to NextGate-1};
    { Doppeleintrge eliminieren }
    g := 1;
    while g < NextGate do begin
      g2 := g;
      while (g2 < NextGate) and (GateBox[g] = GateBox[g2]) and (GateCost[g] =
          GateCost[g2]) and (GateWay^[g].GateDist = GateWay^[g2].GateDist)
          and (GateWay^[g].GateName = GateWay^[g2].GateName) do
        Inc(g2);
      if g2 - g > 1 then begin { doppelte }
        Move(GateWay^[g2], GateWay^[g + 1], (NextGate - g2) * SizeOf(GateWay^[
            g2]));
        Move(GateCost[g2], GateCost[g + 1], (NextGate - g2) * SizeOf(GateCost[
            g2]));
        Move(GateBox[g2], GateBox[g + 1], (NextGate - g2) * SizeOf(GateBox[g2]
            ));
        Dec(NextGate, g2 - g - 1);
      end;
      Inc(g);
    end {while g<NextGate};
    g := 1;
    while g < NextGate do begin
      g2 := g + 1;
      while (g2 < NextGate) do begin
        if GateWay^[g].GateName = GateWay^[g2].GateName then begin
          Move(GateWay^[g2 + 1], GateWay^[g2], (NextGate - g2) * SizeOf(
              GateWay^[g2]));
          Move(GateCost[g2 + 1], GateCost[g2], (NextGate - g2) * SizeOf(
              GateCost[g2]));
          Move(GateBox[g2 + 1], GateBox[g2], (NextGate - g2) * SizeOf(
              GateBox[g2]));
          Dec(NextGate);
        end
        else
          Inc(g2);
      end {while(g2<NextGate)};
      Inc(g);
    end {while g<NextGate};
  end {CalcGateways};

  procedure FixNummern;
  var
    OurNum          : string;
    i               : Box;
    j,
    k,
    l               : Byte;
  begin
    WriteLn(Log, '[FixNummern]');
    OurNum := Boxen^[ThisBox].Nummer1;
    for i := 1 to MaxBoxen do
      if Known[i] then
        with Boxen^[i] do begin
          if i <> ThisBox then begin
            j := 1;
            l := 0;
            k := 0;
            while (OurNum[j] = Nummer1[j + l]) and (j < Length(OurNum)) and (j
                + l < Length(Nummer1)) do begin
              if OurNum[j] = '-' then
                k := j;
              Inc(j);
            end;
            if k > 0 then
              Delete(Nummer1, l + 1, k);
          end {if i<>ThisBox};
        end {with Boxen^[i]};
  end {FixNummern};

  procedure WriteKommTypNames(var O: text; num: string; komm: tConnectSet);
  var
    i: integer;
    t: tConnectTypes;
    isAnalog,
    isDigital: boolean;
  const
    analog: tConnectSet = [cAnalogV32bis, cAnalogV34, cAnalogZyX, cAnalogPEP, cAnalogHST ];
    digital: tConnectSet = [cIsdnX75];
  begin
    isAnalog := analog * komm <> [];
    isDigital := digital * komm <> [];

    if isAnalog and isDigital then
      Writeln(O, 'Analog & ISDN')
    else if isDigital then
      WriteLn(O, 'ISDN X.75')
    else if isAnalog then
      Writeln(O, 'Analog')
    else
      Writeln(O, 'Unbekannt');
  end;

  procedure MakeSBL;
  var
    SBL             : Text;
    b               : Box;
    num             : string;
  begin
    WriteLn(Log, '[MakeSBL]');
    if TestConfigOnly then
      Exit;
    Assign(SBL, MausNetPath + 'SBoxList.Inf');
    Rewrite(SBL);
    WriteLn(SBL, 'Nr. Krzel   Name                           Nummer               Komm.-Typ');
    WriteLn(SBL, '--- -------- ------------------------------ -------------------- -------------');
                 {123 12345678 123456789012345678901234567890 12345678901234567890 1234567890123}
    for b := 1 to MaxBoxen do
      if Known[TelSort[b]] then
        with Boxen^[TelSort[b]] do begin
          Write(SBL, TelSort[b]: 3, ' ', Kuerzel, '': 9 - Length(Kuerzel));
          Write(SBL, Name, '': 31 - Length(Name));
          if TelGeheim[1][TelSort[b]] then
            num := '[nicht ffentlich]'
          else
            num := Nummer1;
          Write(SBL, num, '':21-Length(num));
          WriteKommTypNames(SBL, num, ConnTyp[1]);
          { zweite Nummer: }
          if (Nummer2 <> '') and (not TelGeheim[2][TelSort[b]]) then begin
            Write(SBL, '': 44);
            Write(SBL, Nummer2, '':21-Length(Nummer2));
            WriteKommTypNames(SBL, Nummer2, ConnTyp[2]);
          end;
        end {with Boxen^[TelSort[b]]};
    Close(SBL);
  end {MakeSBL};

  procedure MakeITB;
  var
    b               : Box;
    ITB             : Text;
    g               : Gateway1;
    Exp             : Text;
    l               : string;

    procedure MakeITBNummer(port: integer; nr: string; geheim: boolean; conn: tConnectSet);
    var
      t,p: char;
      i: integer;
      c: tConnectTypes;
    begin
      t := 'T';
      p := 'P';
      if geheim then begin
        t:='t';
        p:='p';
      end;

      if (port=1) then
        WriteLn(ITB, t, nr);

      WriteLn(ITB, p, '#', port);
      WriteLn(ITB, p, t, nr);
      if (conn <> []) then begin
        Write(ITB, p, 'K');
        for c:=low(tConnectTypes) to high(tConnectTypes) do
          if c in conn then
            Write(ITB, copy(cConnectTags,ord(c)+1,1));
        WriteLn(ITB);
      end;
    end;

  begin
    WriteLn(Log, '[MakeITB]');
    if TestConfigOnly then
      Exit;
    Assign(ITB, MausNetPath + 'TBoxList.Inf');
    Rewrite(ITB);
    for b := 1 to MaxBoxen do
      if Known[b] then begin
        if IO2(OpenText(Exp, fmReadIgnore, MakeNameNeu(b) + '.EXP')) then begin
          while not Eof(Exp) do begin
            ReadLn(Exp, l);
            WriteLn(ITB, l);
          end;
          Close(Exp);
        end
        else
          with Boxen^[b] do begin
            WriteLn(ITB, '*', Kuerzel);
            WriteLn(ITB, '#', b);
            WriteLn(ITB, 'N', Name);
            if Known[OnTop[b]] then
              WriteLn(ITB, '-', Boxen^[OnTop[b]].Kuerzel);
            MakeITBNummer(1, Nummer1, TelGeheim[1][b], ConnTyp[1]);
            MakeITBNummer(2, Nummer2, TelGeheim[2][b], ConnTyp[2]);
            for g := 1 to NextGate do
              if GateBox[g] = b then
                WriteLn(ITB, 'G', GateWay^[g].GateName);
          end {with Boxen^[b]};
      end {if Known[b]};
    Close(ITB);
  end {MakeITB};

  procedure SortBoxes;
  var
    i               : Box;
  begin
    WriteLn(Log, '[SortBoxes]');
    for i := 0 to DefBoxenMax do
      BoxSort[i] := i;
    TelSort := BoxSort;
    QSort(1, MaxBoxen, qscBox, qssBox);
    QSort(1, MaxBoxen, qscTel, qssTel);
  end;

begin {ComputeConfig}
  if not TestConfigOnly then begin
    Close(LBL);
    Close(SBI);
  end;
  SortBoxes;
  MakeITB;
  { Files fr InfoMen }
  MakeSBL;
  FixNetwork;
  GenNetzPlan;
  CalcLevels;
  CalcGruppen;
  CalcGateways;
  FixNummern;
  if Debugging then
    GenNetMap;
end {ComputeConfig};

procedure ConfigThis;

  procedure doGBL(var O: text; exclSecret: boolean);
  var
    g:  NetzGruppe;
    b, b0: Box;
    Col: integer;
  const
    w = KuerzelLen;
  begin
    for g := 0 to DefGrpMax do begin
      with GrpInfo[GrpNSort[g]]^ do
        if (Name <> '') and not (Secret and exclSecret) then begin
          Write(O, GrpNSort[g]: 3, ' ');
          if Length(Name) < 11 then
            Write(O, AlignL(Name, 10))
          else
            Write(O, Name);
          WriteLn(O, ' ', UserInfo);
          if BoxListe <> [] then begin
            Write(O, '': 3);
            Col := 3;
            for b0 := 1 to MaxBoxen do begin
              b := BoxSort[b0];
              if Known[b] then begin
                if Col > MaxLineLen - (w+2) then begin
                  WriteLn(O);
                  Write(O, '': 3);
                  Col := 3;
                end;
                if IsInGSet(GrpNSort[g], WillGruppe^[b]) then
                  Write(O, ' ', AlignL(Boxen^[b].Kuerzel, w))
                else
                  Write(O, '': w+1);
                Inc(Col, w+1);
              end {if Known[b]};
            end {for b0:=1 to MaxBoxen};
            WriteLn(O);
          end {if BoxListe<>[]};
        end {if(Name<>'')and not Secret};
    end {for g:=0 to DefGrpMax};

  end;

  procedure ConfigGrps;
  var
    g               : NetzGruppe;
    lg2             : Integer;     (* Wird im Fehlerfall Negativ! *)
    lg,
    grp,
    x               : tGruppe;
    b0,
    b,
    b2,
    b3              : Box;
    db              : DirectBox;
    Col,
    c,
    c2              : Integer;
    GBL,
    GBLs,
    SGL             : Text;
    BP              : PBenutzerprofil;
    BoxSet          : set of Box;
    newpath,
    oldpath         : DirStr;
    s               : string;
    i               : Integer;
  const
    w               = KuerzelLen;
  begin {ConfigGrps}
    WriteLn(Log, '[ConfigGrps]');
    MassenUpdate := True;
    for g := 0 to DefGrpMax do
      GrpNSort[g] := g;
    GrpDSort := GrpNSort;
    QSort(0, DefGrpMax, qscGrpN, qssGrpN);
    QSort(0, DefGrpMax, qscGrpD, qssGrpD);
    WriteLn(Log, '[... init]');
    for lg := 1 to GrpMax do begin { fr alle, Netzflag lschen }
      GetGruppeDirect(lg);
      Gruppe.IstImNetz := False;
      PutGruppeDirect(lg);
    end;
    if TestConfigOnly then begin
      Assign(GBL, 'nul');
      Assign(GBLs, 'nul');
    end
    else begin
      Assign(GBL, MausNetPath + 'GBoxList.Inf');
      Assign(GBLs, MausNetPath + 'GBoxLisG.Inf');
    end;
    Rewrite(GBL);
    Rewrite(GBLs);
    Assign(SGL, MausNetPath + 'SysOpGrp.Inf');
    Rewrite(SGL);
    BoxSet := [];
    for b := 1 to MaxBoxen do begin
      if Known[b] then
        include(BoxSet, b);
    end;
    WriteLn(Log, '[... bestellen]');
    for g := 0 to DefGrpMax do begin
      if GrpInfo[g]^.Name <> '' then begin
        Write(Log, g, ' = ', GrpInfo[g]^.Name);
        WriteLnLog;
      end;
      if IsInGSet(g, WillGruppe^[ThisBox]) then begin
        lg := GruppenNummer(g, '', GrpInfo[g]^.Name);
        if lg = NoGrp then
          Warn('Gruppe nicht gefunden: ' + GrpInfo[g]^.Name);
        if not (ThisBox in GrpInfo[g]^.BoxListe) then begin
          Warn('Gruppe nicht (mehr) fr uns erlaubt: ' + GrpInfo[g]^.Name);
          if not TestConfigOnly then begin
            Warn(' - Gruppe wird abbestellt.');
            DelGSet(g, WillGruppe^[ThisBox]);
            PutBoxGrpRec(ThisBox, WillGruppe^[ThisBox]);
            if (lg <> NoGrp) then begin
              GetGruppeDirect(lg);
              Gruppe.NetGrpNum := 0;
              PutGruppeDirect(lg);
            end;
          end;
          lg := NoGrp;
        end;
        if lg <> NoGrp then begin
          { es gibt eine Gruppe dieses Namens }
          GetGruppeDirect(lg);
        end;
        if (lg <> NoGrp) and (Gruppe.NetGrpNum<>0) then begin
          { Nur wenn sie vernetzt ist! }
          Gruppe.UserInfo := GrpInfo[g]^.UserInfo;
          Gruppe.OriginNet := GrpInfo[g]^.OriginNet;
          Gruppe.GrpLanguage := GrpInfo[g]^.GrpLanguage;
          Gruppe.IstImNetz := True;
          Gruppe.NetGrpNum := g;
          if Gruppe.Name <> GrpInfo[g]^.Name then begin
            { langer/neuer Gruppenname: *mu* gleich Netzname sein! }
            Warn('Gruppe "' + Gruppe.Name + '" mu "' + GrpInfo[g]^.Name +
                '" heien - wird umbenannt');
            lg2 := FindGrpName(GrpInfo[g]^.Name);
            if (lg2 <> -1) and (lg2 <> lg) then begin
              Write(Log, bell, bell, bell);
              Warn('Argh! Lokale Gruppe "' + GrpInfo[g]^.Name + '" existiert bereits!');
              Warn('Bitte die lokale Gruppe lschen oder umbenennen.');
            end
            else begin
              StoreGroupRename(Gruppe.Name, GrpInfo[g]^.Name, Gruppe.Flags.
                  Geheim);
              if (UpStr(Gruppe.Name) <> UpStr(GrpInfo[g]^.Name)) and
                 Gruppe.Flags.HatNenProgteil then begin
                SysopLog('>> Die Gruppe hatte einen Programmteil!');
                SysopLog('>>');
                SysopLog('>> Der Programmteil wurde deaktiviert, da sich der Default-Name fr das');
                SysopLog('>> Directory und der m7com.cfg-Variablenname dadurch gendert haben.');
                SysopLog('>> Bitte neu einrichten und umkopieren!');
                Gruppe.Flags.HatNenProgteil := False;
              end;
              Gruppe.Name := GrpInfo[g]^.Name;
            end;
          end {if((GrpInfo[g]^.Name<>GrpInfo[g]^....};
          PutGruppeDirect(lg);
          if GrpInfo[g]^.Secret then begin
            MakeSecret(lg, False);
            if not SecretOk(lg) then begin
              lg := NoGrp;
              Warn('Nicht ausreichend geheim - nicht konfiguriert');
            end;
          end
          else
            if IsSecret(lg) then
              Warn('Warnung: NetzGruppe ' + IntToStr(g) + ' (' + GrpInfo[g]^.
                  Name + ') ist nicht geheim, wohl aber die lokale Gruppe ' +
                  GruppenName(lg));
        end {if lg<>NoGrp};
        GrpToNetGrp[lg] := g;
        NetGrpInfo[g]^.ToGrp := lg;
        NetGrpInfo[g]^.ToBox[Boxen^[ThisBox].Direct] := ImNetz;
      end {if IsInGSet(g,WillGruppe^[ThisBox])};
      for b := 1 to MaxBoxen do
        if Known[b] then
          if IsInGSet(g, BekommtGruppe^[b]) then begin
            for b2 := 1 to MaxBoxen do
              if Known[b2] then
                if IsInGSet(g, BekommtGruppe^[b2]) then
                  NetGrpInfo[g]^.ToBox[Boxen^[b].Direct] := ImNetz;
          end;
    end {for g:=0 to DefGrpMax};
    WriteLn(Log, '[... GBL]');
    doGBL(GBL, true);
    doGBL(GBLs, false);
    WriteLn(Log, '[... SGL]');
    for g := 0 to DefGrpMax do begin
      with GrpInfo[GrpDSort[g]]^ do
        if Name <> '' then begin
          WriteLn(SGL, GrpDSort[g]: 3, ' ', Name, '': 40 - Length(Name),
              ' zuletzt editiert ', DatumSeit(LastEdit));
          WriteLn(SGL, 'User-Info:  ', UserInfo);
          WriteLn(SGL, 'Sysop-Info: ', SysOpInfo);
          WriteLn(SGL, 'Ursprungsnetz: ', NetworkName(OriginNet));
          WriteLn(SGL, 'Sprache: ', LanguageName(GrpLanguage));
          if BoxListe * BoxSet = BoxSet then
            WriteLn(SGL, 'Erlaubt fr alle Boxen')
          else begin
            Write(SGL, 'Erlaubt fr:');
            Col := 12;
            for b0 := 1 to DefBoxenMax do begin
              b := BoxSort[b0];
              if b in BoxListe * BoxSet then begin
                if Col > MaxLineLen - (w+2) then begin
                  WriteLn(SGL);
                  Write(SGL, '': 12);
                  Col := 12;
                end;
                Write(SGL, ' ', AlignL(Boxen^[b].Kuerzel, w));
                Inc(Col, w+1);
              end;
            end {for b0:=1 to DefBoxenMax};
            WriteLn(SGL);
          end {if BoxListe*BoxSet=BoxSet else};
          if Secret then
            WriteLn(SGL, '(Geheim)');
          WriteLn(SGL);
        end {if Name<>''};
    end {for g:=0 to DefGrpMax};
    ResetUserLong;
    WriteLn(Log, '[... Fremdbox-User]');
    for b := 1 to MaxBoxen do begin
      if Known[b] then
        if Boxen^[b].FremdBox and (OnTop[b] = ThisBox) then begin
          WriteLn(Log, 'Konfiguriere Fremdbox ', Boxen^[b].Name);
          BP := GetUserLongDirect(- 1000 - b);
          with BP^ do begin
            if Status = 'X' then
              Status := 'N';
            { InitGSet(ShowGruppe); -- warum?? }
            for g := 0 to DefGrpMax do
              if IsInGSet(g, BekommtGruppe^[b]) then begin
                lg := GruppenNummer(g, '', GrpInfo[g]^.Name);
                if lg = NoGrp then
                  Warn('Gruppe nicht gefunden: ' + GrpInfo[g]^.Name);
                if lg <> NoGrp then begin
                  SetGSet(lg, ShowGruppe);
                  if not (b in GrpInfo[g]^.BoxListe) then begin
                    Warn('Gruppe ' + GrpInfo[g]^.Name + ' nicht erlaubt - wird abbestellt.');
                    DelGSet(lg, ShowGruppe);
                    GetBoxGrpRec(b, WillGruppe^[b]);
                    DelGSet(g, WillGruppe^[b]);
                    PutBoxGrpRec(b, WillGruppe^[b]);
                  end;
                end;
              end {if IsInGSet(g, BekommtGruppe^[b])};
            InGruppe := ShowGruppe;
            Header := Boxen^[b].Name + ' ' + Boxen^[b].Nummer1;
            Vorname := 'MAUS';
            Name := Boxen^[b].Kuerzel;
          end {with BP^};
          PutUserLong;
          for g:=0 to DefGrpMax do begin
            if IsInGSet(g, BP^.ShowGruppe) then begin
              GetGruppeDirect(g);
              with Gruppe do begin
                Write(Log, 'bekommt Gruppe ', Name);
                if NetGrpNum<>0 then
                  Write(Log, ' (Netz #', NetGrpNum, ')')
                else
                  Write(Log, ' (lokal)');
                WriteLnLog;
              end;
            end;
          end;
        end {if Boxen^[b].FremdBox and(OnTop[b]...};
    end {for b:=1 to MaxBoxen};
    CloseUserLong;
    Close(GBL);
    Close(GBLs);
    Close(SGL);
    MassenUpdate := False;
    CloseGrpFile;
  end {ConfigGrps};

  procedure ConfigLinks;
  var
    b               : Box;
    done            : Boolean;
  begin
    WriteLn(Log, '[ConfigLinks]');
    MaxDirectBox := 0;
    if OnTop[ThisBox] <> 0 then
      Boxen^[OnTop[ThisBox]].UeberMir := True;
    for b := 1 to MaxBoxen do
      if Known[b] then begin
        if OnTop[b] = ThisBox then
          Boxen^[b].UnterMir := True;
        Boxen^[b].Ebene := Level[b];
      end;
    for b := 1 to MaxBoxen do
      if Known[b] then
        if Boxen^[b].UeberMir or Boxen^[b].UnterMir then begin
          Boxen^[b].ViaBox := b;
          Inc(MaxDirectBox);
          if MaxDirectBox > DefDirectBoxMax then begin
            FatalError('Zu viele direkt angeschlossene Boxen!!');
          end;
          Boxen^[b].Direct := MaxDirectBox;
          DirectToBox[MaxDirectBox] := b;
        end;
    done := False;
    while not done do begin
      done := True;
      for b := 1 to MaxBoxen do
        if Known[b] then begin
          if Boxen^[b].FremdBox and (OnTop[b] <> 0) and Boxen^[OnTop[b]].
              UnterMir and Boxen^[OnTop[b]].FremdBox and not Boxen^[b].
              UnterMir then begin
            Boxen^[b].UnterMir := True;
            done := False;
          end;
        end;
    end {while not done};
    repeat
      done := True;
      for b := 1 to MaxBoxen do
        if Known[b] and (OnTop[b] <> 0) { sonst chaos } then begin
          if Boxen^[b].ViaBox = 0 then begin
            if Boxen^[OnTop[b]].ViaBox <> 0 then begin
              Boxen^[b].ViaBox := Boxen^[OnTop[b]].ViaBox; done := False;
            end;
          end
          else
            if Boxen^[OnTop[b]].ViaBox = 0 then begin
              Boxen^[OnTop[b]].ViaBox := Boxen^[b].ViaBox; done := False;
            end;
        end;
    until done;
    Boxen^[ThisBox].ViaBox := 0;
    for b := 1 to MaxBoxen do
      if Known[b] and (b <> ThisBox) then
        if Boxen^[b].Direct = 0 then
          Boxen^[b].Direct := Boxen^[Boxen^[b].ViaBox].Direct;
    Boxen^[ThisBox].Direct := 0;
    DirectToBox[0] := ThisBox;
    for b := 1 to MaxDirectBox do
      WriteLn(Log, 'Direktverbindung #', b, ' ist ', Boxen^[DirectToBox[b]].
          Kuerzel);
  end {ConfigLinks};

  procedure ConfigGates;
  var
    w               : Gateway1;
    BP              : PBenutzerprofil;
  begin
    WriteLn(Log, '[ConfigGates]');
    for w := 1 to NextGate - 1 do
      with GateWay^[w] do begin
        Write(Log, 'Gateway #', w, ' ', GateName, ' [', GateCost[w] - Longint(
            1), '] ');
        if GateBox[w] = ThisBox then
          Write(Log, ' lokal ', GateDist)
        else begin
          Write(Log, ' in ', GateBox[w], '(', Boxen^[GateBox[w]].Kuerzel,
              ') als ', GateDist, ' via ', Boxen^[GateBox[w]].ViaBox);
          GateDist := GateBox[w];
        end;
        WriteLnLog;
      end;
    ResetUserLong;
    for w := 1 to NextGate - 1 do
      if GateBox[w] = ThisBox then
        with GateWay^[w] do
          if GateDist < NetzMitt then begin
            WriteLn(Log, 'Konfiguriere Gateway ', GateName);
            BP := GetUserLongDirect(GateDist);
            with BP^ do begin
              if Status = 'X' then
                Status := 'N';
              if Header = '' then
                Header := 'Gate ' + IntToStr(GateDist);
              Vorname := 'GATE';
              if (Name = '') or (Name = '?') then
                Name := GateName;
            end;
            PutUserLong;
          end {if GateDist<NetzMitt};
    CloseUserLong;
  end {ConfigGates};

begin {ConfigThis}
  ConfigLinks;
  ConfigGrps;
  ConfigGates;
  NetConfigSchreiben(TestConfigOnly);
  if (not TestConfigOnly) and (WriteSuccessMsg) then
    SysOpLog('Konfiguration generiert.');
  SysopProtExit('MNConfig', 'MNConfig-Report');
end {ConfigThis};

var
  i               : Word;
  Dat             : string[8];
begin {mnConfMake}
  UnitVersion('mnConfMake', 'kh, 20.02.95');
  DrawConfigLine('MNConfig - Globale Einstellungen');
  New(Boxen);
  New(OldBoxen);
  New(WillGruppe);
  New(WollteGruppe);
  New(BekommtGruppe);
  New(BekamGruppe);
  SysopProtInit;
  New(TelTrans[1]);
  GetConfigString(TelTrans[1]^.Find, 'TelTrans1Find', '+');
  GetConfigString(TelTrans[1]^.Repl, 'TelTrans1Repl', '00');
  New(TelTrans[2]);
  GetConfigString(TelTrans[2]^.Find, 'TelTrans2Find', '+49-');
  GetConfigString(TelTrans[2]^.Repl, 'TelTrans2Repl', '0');
  for i := 3 to TelTransMax do begin
    New(TelTrans[i]);
    with TelTrans[i]^ do begin
      GetConfigString(Find, 'TelTrans' + IntToStr(i) + 'Find', #255);
      GetConfigString(Repl, 'TelTrans' + IntToStr(i) + 'Repl', '');
    end;
  end;
  WriteSuccessMsg := GetConfigBoolean('WriteSuccessMsg', false);
end {mnConfMake}.
